/*  $Id: SymTab.cc,v 1.17 1999/07/16 20:18:51 shaggy Exp $
 *      Symbol Table support
 *      
 *      relative path: lib/symtab
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define IMPLEMENTATION_SymTab
#include <stdx1.h>
#include <SymTab>
#include <new>

// XXX - not thread safe


// obj_t Type_vSymTab;

// note: the cache mechanism here is currently broken
//   ^ is this note still valid?  hasn't it been fixed?
//	I can't remember, and don't really have time
//	to check right now...

SymTab::SymTab(vStrTab *strtab) 
{
	_strtab = strtab;
	_first = 0;
	_cache = 0;
	_inum = 0;
	_inext = 0;
	_parent = 0;
	_data = 0;
}

SymTab::SymTab(vSymTab *parent)
{
	new(this)SymTab(parent->strtab());
	_parent = parent;
}

SymTab::SymTab(vSymTab *parent, void *data)
{
	new(this)SymTab(parent);
	_data = data;
}


SymTab*
SymTab::transplant()
{
	SymTab *s = new SymTab(_parent, _data);
	s->_first = _first;
	s->_cache = _cache;
	new(this)SymTab(_parent, _data);
	return s;
}

SymTab::~SymTab()
{
	_SymEntry *sp = _first;

	while(sp) {
		_SymEntry *spnext = sp->next;
		// XXX - should probably try to free the value...
		// but we don't really know its type or any way to 
		// access it...

		delete sp;
		sp = spnext;
	}
}


vSymTab*
SymTab::parent() const
{
	return _parent;
}

bool
SymTab::parent(vSymTab *p)
{
	_parent = p;
	return true;
}

void*
SymTab::data() const
{
	return _data;
}

bool
SymTab::data(void *d)
{
	// if(_data)
	//	return false;
	// else {
		_data = d;
		return true;
	// }
}

vStrTab*
SymTab::strtab() const
{
	return _strtab;
}


bool SymTab::get(Id id, Val &val) const
{
	_SymEntry *psp = 0;
	_SymEntry *sp = _inext;
	if(!sp || sp->id != id)
		for(sp = _cache; sp && sp->id != id; sp = sp->cnext)
			psp = sp;
	if(!sp)
		return false;
	if(psp) {
		psp->cnext = sp->cnext;
		sp->cnext = _cache;
		_cache = sp;
	}
	val = sp->val;
	return true;
}	

bool SymTab::set(Id id, Val val)
{
	_SymEntry *psp = 0;
	_SymEntry *sp = _inext;
	if(!sp || sp->id != id)
		for(sp = _cache; sp && sp->id != id; sp = sp->cnext)
			psp = sp;
	if(!sp)
		return false;
	if(psp) {
		psp->cnext = sp->cnext;
		sp->cnext = _cache;
		_cache = sp;
	}
	sp->val = val;
	return true;
}

bool SymTab::add(Id id, Val val)
{
	Val tval;
	if(get(id, tval))
		return false;
	_SymEntry *sp = new _SymEntry;
	sp->id = id;
	sp->val = val;
	sp->next = _first;
	sp->cnext = _cache;
	_first = sp;
	_cache = sp;
	return true;
}


bool SymTab::del(Id id)
{
	_SymEntry *psp = 0;
	_SymEntry *sp = _inext;
	if(sp && sp->id == id) {
		_inext = _inext->cnext;
		++_inum;
	} else {
		for(sp = _cache; sp && sp->id != id; sp = sp->cnext)
			psp = sp;
		if(!sp)
			return false;
		if(psp) 
			psp->cnext = sp->cnext;
		else
			_cache = sp->cnext;
	}
	psp = _first;
	if(psp == sp)
		_first = sp->next;
	else {
		while(psp->next != sp)
			psp = psp->next;
		psp->next = sp->next;
	}
	delete sp;
	return true;
}


// Although index() will allow random access, only forward sequential
// access will be efficient
// Also, as long as the cached entry is used, other threads inserting
// or deleting elements will not cause dropouts 
// but if the cache isn't used, the ordering could change, and
// dropouts or repeats could occur

SymTab::Id SymTab::index(uint index) const
{
	if(!index || index < _inum) {
		_inum = 0;
		_inext = _first;
	}
	while(_inext && _inum < index) {
		_inext = _inext->next;
		++_inum;
	}
	return _inext ? _inext->id : 0;
}	

