/*  $Id: genc2a.cc,v 1.7 1999/07/16 20:18:52 shaggy Exp $
 *      Dynamic Encapsulator of C++ Objects
 *      
 *      relative path: util
 *  
 *  Copyright (c) 1996-1998 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define genc2a_VER_MAJOR	0
#define genc2a_VER_MINOR	3
#define genc2a_VER_PATCH	0
#define genc2a_VER_STR	"0.3"
#define genc2a_VER_DATE	"August 17, 1998"
#define genc2a_VER_NDATE	"08/17/98"

#include <stdx1.h>
#include <libc_posix.h>
#include <ldefines>
#include <defines>
#include <assert.h>

#include <CObj>
#include <CLex>
#include <StrTab>
#include <SymTab>
#include <ParsePath>
#include <vCAction>
#include <CParse>
#include <CRep>
#include <COut>
#include <CAction>


bool debugaction = false;

// lexical shortcuts:
typedef vStrTab::Id Id;


class Gen : public CAction {
private:
	int _input_fd;
	char _buf[4096];

public:
	Gen(CLex &clp, vStrTab &strtab, vSymTab &symtab);

	Decl* declfind(csz *name, Scope sc = 0);
	bool open_input(csz *fname = 0);
	void close_input();
	void get_input(bool split_token);
	void abort_exit();
	void internal_error_exit();
	void veprintf(csz *fmt, va_list args);

	void printdecls(Decl *d, csz *pfmt, csz *sid, csz *sn);
};


Gen::Gen(CLex &clp, vStrTab &strtab, vSymTab &symtab)
	 : CAction(clp, strtab, symtab)
{
}


Decl*
Gen::declfind(csz *name, Scope sc)
{
	if(!sc)
		sc = global();
	if(csz *s = strstr(name, "::")) {
		char buf[MAXIDLEN];
		strncpy(buf, name, s-name);
		buf[s-name] = '\0';
		Decl *d = sc.tget(id(buf));
		if(!d) {
			error("`%s' not found", buf);
			return 0;
		}
		Scope nsc = getaggrscope(d);
		if(!nsc) {
			error("`%s' not a scope", buf);
			return 0;
		}
		return declfind(s+2, nsc);
	}	
	Decl *d = sc.tfind(id(name));
	if(!d) 
		error("`%s' not declared", name);
	return d;
}


/////////////////////////////////////////////////////////////////////////////
// I/O functionality

void
Gen::veprintf(csz *fmt, va_list args)
{
	vfprintf(stderr, fmt, args);
}



void
Gen::abort_exit()
{
	exit(1);
}


void
Gen::internal_error_exit()
{
	mprintf("", "Please send a full bug report to `shaggy@csh.rit.edu'");
	mprintf("", "(please read the BUGS file first for details)");
	exit(2);
}


bool
Gen::open_input(csz *fname)
{
	_input_fd = fname ? open(fname, O_RDONLY) : 0;
	_clex.source(fname ? fname : "");
	return _input_fd >= 0;
}


void
Gen::close_input()
{
	if(_input_fd)
		close(_input_fd);
}


void
Gen::get_input(bool split_token)
{
	if(isatty(_input_fd) && !split_token) 
		write(1, (complete() ? "> " : "? "), 2);
	int n = read(_input_fd, _buf, sizeof _buf);
	if(n <= 0)
		_clex.set_eof();
	else
		_clex.set_src(_buf, n);
}
 

static void
idstring(sz *s)
{
	while(*s) {
		if(*s != '_' && (*s < 'a' || *s > 'z')
				&& (*s < 'A' || *s > 'Z')
				&& (*s < '0' || *s > '9'))
			*s = '_';
		++s;
	}
}

void
Gen::printdecls(Decl *d, csz *pfmt, csz *sid, csz *sn)
{
	for(Decl *dm = d->afirst(); dm; dm = dm->next()) {
		if(dm->access() != Decl::ac_public)
			continue;
		if(dm->dtype() == Decl::dt_enum)
			printdecls(dm, pfmt, sid, sn);
		if(!dm->field() && dm->dtype() != Decl::dt_enumval
				&& dm->storage() != Decl::sc_constant)
			continue;
		if(dm->anon() || dm->id() == ID_ANON)
			continue;
		csz *mn = cstr(dm->id());
		char mid[MAXIDLEN];
		strcpy(mid, mn);
		idstring(mid);
		printf(pfmt, sid, mid);
		if(dm->field())
			printf("(int)&((%s*)0)->%s);\n", sn, mn);
		else
			printf("%s::%s);\n", strstr(sn, " ")+1, mn);
	}
}

int
main(int argc, sz **argv)
{
	char strbuf[4096];
	char bufbuf[4096];
	char namebuf[1024];

	CLex lex(strbuf, sizeof(strbuf));
	lex.set_buf(bufbuf, sizeof(bufbuf));
	lex.set_name(namebuf, sizeof(namebuf));
	lex.parse_keywords();
	lex.parse_space(false).parse_indent(false);

	StrTab strtab;
	SymTab symtab(&strtab);
	Decl gdecl(0,0);
	Scope(&symtab).decl(&gdecl);
	Gen gen(lex, strtab, symtab);
	CParse cparse(lex, gen);
	bool first = true;
	cparse.savecode(true);
	bool nolang = true;

	int argnum = 0;
	while(++argnum < argc) {
		csz *a = argv[argnum];

		if(a[0] == '-' && a[1]) {	// it's an option

			// Overall Options
			if(strcmp(a, "-x") == 0) {
				a = argv[++argnum];
				if(!a) {
					fprintf(stderr, "genc2a: missing arg\n");
					exit(1);
				}
				nolang = false;
				if(strcmp(a, "c") == 0)
					gen.parse_c();
				else if(strcmp(a, "c-hdr") == 0)
					gen.parse_h();
				else if(strcmp(a, "c-header") == 0)
					gen.parse_h();
				else if(strcmp(a, "c++") == 0)
					gen.parse_cxx();
				else if(strcmp(a, "c++hdr") == 0)
					gen.parse_hxx();
				else if(strcmp(a, "dc++") == 0)
					gen.parse_dcxx();
				else if(strcmp(a, "dc++hdr") == 0)
					gen.parse_dhxx();
				else if(strcmp(a, "none") == 0)
					nolang = true;
				else {
					fprintf(stderr,
					    "genc2a: unknown language `%s'\n",
						a);
					exit(1);
				}	
			}
			else if(strcmp(a, "-v") == 0) {
				printf("genc2a " genc2a_VER_STR " ("
					genc2a_VER_DATE ")\n");
				return 0;
			}

			else
				fprintf(stderr,
					"genc2a: unknown option `%s'\n", a);
			continue;
		}

		if(first) {
			if(nolang) 
			    gen.parse_dcxx();
			if(!gen.open_input(0)) {
				perror("stdin");
				break;
			}
			if(!cparse.parse()) 
				gen.mprintf("", "cannot recover from previous errors");
			if(gen.warnings() > 0)
				gen.mprintf("", "%d warning%s",
					gen.warnings(), gen.warnings() == 1 ? "" : "s");
			if(gen.errors() > 0) {
				gen.mprintf("", "%d error%s",
					gen.errors(), gen.errors() == 1 ? "" : "s");
				return 1;
			} 
			gen.close_input();
			printf("/* this file was auto-generated by genc2a "
				genc2a_VER_STR " from the command:\n *\t");
			for(int i=0; i<argc; i++)
				printf("%s ", argv[i]);
			printf("\n */\n\n");
			printf("#ifdef __cplusplus\n"
				"\textern \"C\" int printf(const char *, ...);\n"
				"#else\n"
				"\textern int printf(const char *, ...);\n"
				"#endif\n\n");
			printf("int\nlog2(int x)\n{\n"
				"\tint p=0;\n"
				"\twhile((x >>= 1) != 0) ++p;\n"
				"\treturn p;\n"
				"}\n\n");
			printf("int\nmain()\n{\n");
			first = false;
		} 
		Decl *d = gen.declfind(a);
		if(!d) {
			gen.warning("cannot find aggregate `%s'", a);
			continue;
		}
		if(!d->dtaggr()) {
			gen.error("`%s' is not an aggregate", gen.cstr(d->id()));
			continue;
		}
		char sid[MAXIDLEN], sn[MAXIDLEN];
		csz *pfmt = "\tprintf(\"#define %%s__%%-32s 0x%%04x\\n\", \"%s\", "
				"\"%s\", ";
		gen.ccout.scoped_string(d, sn, false);
		strcpy(sid, sn);
		idstring(sid);
		strpcat(sn, " ");
		strpcat(sn, d->dtname());
		gen.printdecls(d, pfmt, sid, sn);
		printf(pfmt, sid, "sizeof");
		printf("sizeof(%s));\n", sn);
		printf("\tif((1 << log2(sizeof(%s))) == sizeof(%s))\n\t", sn, sn);
		printf(pfmt, sid, "2sizeof");
		printf("log2(sizeof(%s)));\n", sn);
		printf("\tprintf(\"\\n\");\n");
		printf("\n");
	}
	if(first) {
		fprintf(stderr, "usage: genc2a [-x lang] aggregate...\n");
		return 1;
	} 
	printf("\treturn 0;\n");
	printf("}\n");
	return 0;
}

